#include <Arduino.h>
#include <stdlib.h>
#include <stdint.h>
#include <config.h>
#include <DFRobot_ST7687S_Widgets.h>
#include <Adafruit_PN532.h>
#include <imgs.h>
#include "BLEDevice.h"

DFRobot_ST7687S_Widgets tft_left(TFT_LEFT_PIN_CS, TFT_LEFT_PIN_RS, TFT_LEFT_PIN_WR, TFT_LEFT_PIN_LCK);
DFRobot_ST7687S_Widgets tft_right(TFT_RIGHT_PIN_CS, TFT_RIGHT_PIN_RS, TFT_RIGHT_PIN_WR, TFT_RIGHT_PIN_LCK);

bool doConnect = false;
bool connected = false;
bool doScan = false;
bool offroad_deactivated = false;
bool offroad_touch_released = false;

int btn_checked = millis();
int touch_started = 0;
int init_done = 0;
int last_status = 0;
int bike_mode = BIKE_MODE_NORMAL;

static BLEUUID serviceUUID("0000ffe0-0000-1000-8000-00805f9b34fb");
static BLEUUID charUUID("0000ffe1-0000-1000-8000-00805f9b34fb");
static BLERemoteCharacteristic* pRemoteCharacteristic;
static BLEAdvertisedDevice* myDevice;

// TODO: check request length
// uint8_t status_request[] = { 0x66, 0x41, 0x02, 0x05, 0x00, 0xAE, 0x00, 0x00, 0x30, 0xDD, 0xFC, 0x3F, 0xF0, 0xDC, 0xFC, 0x3F, 0x66, 0x10, 0x94, 0xEE, 0x13, 0x00, 0x00, 0x00 };
//uint8_t old_speed = 0;
uint8_t status_request[] = { 0x66, 0x41, 0x02, 0x05, 0x00, 0xAE };
uint8_t last_bike_status[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
uint8_t bike_status[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };

float calc_speed(uint8_t param_0, uint8_t param_1) {
  int rmp = (int) (60.0d / (((double) ((param_0 * 255) + param_1)) / 1000.0d));
  uint8_t tempSpeed = (int) ((((double) rmp) * ((((double) ((float) 27) / 39.37f)) * 3.141592653589793d) / 1000.0d) * 60.0d);
  return tempSpeed;
}

static void notifyCallback(BLERemoteCharacteristic* pBLERemoteCharacteristic, uint8_t* pData, size_t length, bool isNotify) {

  if ( (length == 9) && (pData[0] ==0x66) && (pData[1] ==0x41) && (pData[2] ==0x05) ) {
    for (int i = 0; i < length; i++) {
      last_bike_status[i] = bike_status[i];
      bike_status[i] = pData[i];
    }
  }

  else {
    Serial.print("size: ");
    Serial.print(length);
    Serial.print(" | data: ");
    for (int i = 0; i < length; i++) {
      Serial.print(pData[i], HEX);
      Serial.print(" ");
    }
    Serial.println("");
  }
}

class MyClientCallback : public BLEClientCallbacks {
  void onConnect(BLEClient* pclient) { }
  void onDisconnect(BLEClient* pclient) { connected = false; }
};

class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
  void onResult(BLEAdvertisedDevice advertisedDevice) {
    if (advertisedDevice.getAddress().equals(BLEAddress(RAYVOLT_BIKE))) {
      BLEDevice::getScan()->stop();
      myDevice = new BLEAdvertisedDevice(advertisedDevice);
      doConnect = true;
      doScan = true;
    }
    if (advertisedDevice.getAddress().equals(BLEAddress(RAYVOLT_DUMMY))) {
      BLEDevice::getScan()->stop();
      myDevice = new BLEAdvertisedDevice(advertisedDevice);
      doConnect = true;
      doScan = true;
    }
  }
};

bool connectToServer() {
    BLEClient*  pClient  = BLEDevice::createClient();
    pClient->setClientCallbacks(new MyClientCallback());
    pClient->connect(myDevice);
    BLERemoteService* pRemoteService = pClient->getService(serviceUUID);
    if (pRemoteService == nullptr) {
      pClient->disconnect();
      return false;
    }

    pRemoteCharacteristic = pRemoteService->getCharacteristic(charUUID);
    if (pRemoteCharacteristic == nullptr) {
      pClient->disconnect();
      return false;
    }

    if(pRemoteCharacteristic->canRead()) {
      std::string value = pRemoteCharacteristic->readValue();
    }

    if(pRemoteCharacteristic->canNotify())
      pRemoteCharacteristic->registerForNotify(notifyCallback);

    connected = true;
    return true;
}

bool check_touch() {
  if ( (touchRead(TOUCH_BUTTON) < TOUCH_THRESHOLD) && (touch_started != 0) && millis() - touch_started > TOUCH_CHECK_DURATION ) {
    return true;
  }
  else if ( (touchRead(TOUCH_BUTTON) < TOUCH_THRESHOLD) && touch_started == 0 ) {
    touch_started = millis();
  }
  else if ( !(touchRead(TOUCH_BUTTON) < TOUCH_THRESHOLD) && touch_started != 0 ) {
    touch_started = 0;
  }
  return false;
}

void check_buttons() {
  if ( (digitalRead(BTN_RIGHT_LOWER)) && (millis() - btn_checked > BTN_COOLDOWN) ) {
    btn_checked = millis();

    if (status_request[3] > 2 ) {
      status_request[3]--;
      tft_right.PasDisplay((status_request[3]) - 2);
    }
  }

  else if ( (digitalRead(BTN_RIGHT_CENTER)) && (millis() - btn_checked > BTN_COOLDOWN) ) {
    btn_checked = millis();
  }

  else if ( (digitalRead(BTN_RIGHT_HIGHER)) && (millis() - btn_checked > BTN_COOLDOWN) ) {
    btn_checked = millis();

    if (status_request[3] < 5 ) {
      status_request[3]++;
      tft_right.PasDisplay((status_request[3]) - 2);
    }
  }
}

void process_status() {
//  TODO: ERRORS OverVoltage / Undervoltage / BLE_ERROR
/*
  Serial.print("bike_status");
  for (int i = 0; i < sizeof(bike_status); i++) {
    Serial.print(bike_status[i], BIN);
    Serial.print(" ");
  }
  Serial.println("");
*/
  if( ((last_bike_status[7] >> 0) & 1) != ((bike_status[7] >> 0) & 1) ) {
    if((bike_status[7] >> 0) & 1) {
      Serial.println("rbs error");
      tft_right.drawBmp(RegenBrakeFaultBMP, -30, 20, 25, 25, 1, REGENBRAKE_FAULT_COLOR);
    }
    else {
      tft_right.fillRect(-30, 20, 25, 25, CBC_DISPLAY_BACKCOLOR);
    }
  }
  
  if( ((last_bike_status[7] >> 1) & 1) != ((bike_status[7] >> 1) & 1) ) {
    if((bike_status[7] >> 1) & 1) {
     tft_right.drawBmp(ThrottleFaultBMP, 5, 20, 25, 25, 1, THROTTLE_FAULT_COLOR);
    }
    else {
      tft_right.fillRect(5, 20, 25, 25, CBC_DISPLAY_BACKCOLOR);
    }
  }

  if ( last_bike_status[3] != bike_status[3] ) {
    if (bike_status[3] == 0) {
      tft_right.fillRect(30, -28, 25, 50, CBC_DISPLAY_BACKCOLOR);
      tft_right.drawBmp(Battery0BMP, 30, -28, 25, 50, 1, DISPLAY_RED);
    }
    else if (bike_status[3] == 1) {
      tft_right.fillRect(30, -28, 25, 50, CBC_DISPLAY_BACKCOLOR);
      tft_right.drawBmp(Battery1BMP, 30, -28, 25, 50, 1, DISPLAY_RED);
    }
    else if (bike_status[3] == 2) {
      tft_right.fillRect(30, -28, 25, 50, CBC_DISPLAY_BACKCOLOR);
      tft_right.drawBmp(Battery2BMP, 30, -28, 25, 50, 1, DISPLAY_ORANGE);
    }
    else if (bike_status[3] == 3) {
      tft_right.fillRect(30, -28, 25, 50, CBC_DISPLAY_BACKCOLOR);
      tft_right.drawBmp(Battery3BMP, 30, -28, 25, 50, 1, DISPLAY_ORANGE);
    }
    else if (bike_status[3] == 4) {
      tft_right.fillRect(30, -28, 25, 50, CBC_DISPLAY_BACKCOLOR);
      tft_right.drawBmp(Battery4BMP, 30, -28, 25, 50, 1, DISPLAY_GREEN);
    }
    else if (bike_status[3] == 5) {
      tft_right.fillRect(30, -28, 25, 50, CBC_DISPLAY_BACKCOLOR);
      tft_right.drawBmp(Battery5BMP, 30, -28, 25, 50, 1, DISPLAY_GREEN);
    }
  }
  if ( (last_bike_status[5] != bike_status[5]) || (last_bike_status[6] != bike_status[6]) ) {
    Serial.print("Speed: ");
    Serial.println(calc_speed(bike_status[5], bike_status[6]));
    tft_left.TachoNeedle((int) calc_speed(bike_status[5], bike_status[6]));
  }
}

void setup() {
  Serial.begin(115200);

  while( !(millis() > (TOUCH_CHECK_DURATION * 2)) ) {
    if (check_touch()) {
      Serial.println("Offroad not possible");
      offroad_deactivated = true;
    }
  }

  pinMode(BTN_RIGHT_LOWER, INPUT);
  pinMode(BTN_RIGHT_CENTER, INPUT);
  pinMode(BTN_RIGHT_HIGHER, INPUT);

  tft_left.begin(ST7687S_Rotation_270);
  tft_left.fillScreen(CBC_DISPLAY_BACKCOLOR);
  
  tft_right.begin(ST7687S_Rotation_90);
  tft_right.fillScreen(CBC_DISPLAY_BACKCOLOR);

  tft_right.AmpMeter(50);
  tft_right.PasDisplay((status_request[3]) - 2);

  tft_left.drawNormalTacho();
  tft_left.TachoLabels();
  tft_left.TachoNeedle(35);

  BLEDevice::init("");
  BLEScan* pBLEScan = BLEDevice::getScan();
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setInterval(1349);
  pBLEScan->setWindow(449);
  pBLEScan->setActiveScan(true);
  pBLEScan->start(5, false);

  init_done = millis();
}

void loop() {

  if (check_touch()) {
    if ( (bike_mode == BIKE_MODE_NORMAL) && ((millis() - init_done) < (OFFROAD_WINDOW * 1000)) && !offroad_deactivated) {
      bike_mode = BIKE_MODE_OFFROAD;
      tft_left.drawOffRoadTacho();
      Serial.println("switching to offroad");
    }
    else if ( (bike_mode == BIKE_MODE_OFFROAD) && offroad_touch_released ) {
      offroad_deactivated = true;
      bike_mode = BIKE_MODE_NORMAL;
      tft_left.drawNormalTacho();
      Serial.println("switching to normal");
    }
  }
  else if ( (bike_mode == BIKE_MODE_OFFROAD) && !offroad_touch_released ) {
    Serial.println("offroad_touch_released");
    offroad_touch_released = true;
  }

  if (doConnect == true) {
    if (connectToServer()) { Serial.println("connected"); }
    else { Serial.println("failed to connect"); }
    doConnect = false;
  }

  if (connected) {
    if ( (millis() - last_status) > STATUS_DURATION) {
      last_status = millis();
      Serial.println("sending status_request");
      pRemoteCharacteristic->writeValue(status_request, sizeof(status_request));
    }
    check_buttons();
    process_status();
  }

  else if(doScan){
    Serial.println("disconnected");
//    BLEDevice::getScan()->start(0);  // this is just eample to start scan after disconnect, most likely there is better way to do it in arduino
  }

  delay(10);
}